# Pyramid
from pyramid.config import Configurator

# Websauna
from websauna.system import Initializer
from websauna.system.model.utils import attach_models_to_base_from_module
from websauna.utils.autoevent import after
from websauna.utils.autoevent import bind_events


class AddonInitializer:
    """Configure this addon for websauna.

    * We hook to existing parts of initialization process using ``@after`` aspect advisor

    * For something we don't have a direct join point we just initialize through ``run()``
    """

    def __init__(self, config: Configurator):
        self.config = config

    @after(Initializer.configure_templates)
    def configure_templates(self):
        """Include our package templates folder in Jinja 2 configuration."""

        # Use prepend=False here so that the app can override our templates easily

    @after(Initializer.configure_instrumented_models)
    def configure_instrumented_models(self):
        """Plug models from this addon to Websauna."""
        from . import models
        from websauna.system.model.meta import Base

        # Load all models (if we have any) and attach them to SQLALchemy default base class
        attach_models_to_base_from_module(models, Base)

    def configure_addon_views(self):
        """Configure views for your application.

        Let the config scanner to pick ``@simple_route`` definitions from scanned modules. Alternative you can call ``config.add_route()`` and ``config.add_view()`` here.
        """
        # We override this method, so that we route home to our home screen, not Websauna default one
        from . import views
        self.config.scan(views)

    def run(self):

        # This will make sure our initialization hooks are called later
        bind_events(self.config.registry.initializer, self)

        # Run our custom initialization code which does not have a good hook
        self.configure_addon_views()


def includeme(config: Configurator):
    """Entry point for Websauna main app to include this addon.

    In the Initializer of your app you should have:

        def include_addons(self):
            # ...
            self.config.include("websauna.magiclogin")

    """
    addon_init = AddonInitializer(config)
    addon_init.run()



# Pyramid
import colander
import deform
from pyramid import httpexceptions
from pyramid.settings import aslist

# Websauna
from websauna.magiclogin.requirelogin import get_login_state
from websauna.system.core.route import simple_route
from websauna.system.core.sitemap import include_in_sitemap
from websauna.system.form.schema import CSRFSchema
from websauna.system.form.throttle import throttled_view
from websauna.system.http import Request

from .login import start_email_login
from .login import verify_email_login


class AskEmailSchema(CSRFSchema):
    """Form for getting user email."""

    email = colander.SchemaNode(
        colander.String(),
        title="Email address",
        validator=colander.Email(),
        widget=deform.widget.TextInputWidget(type='email', template="textinput_placeholder", placeholder="yourname@example.com")
    )


@simple_route("/login", route_name="login", renderer='magiclogin/login.html')
@include_in_sitemap(False)
def login(request: Request):
    """Replace the defaut login view with this simplified version."""
    settings = request.registry.settings
    social_logins = aslist(settings.get("websauna.social_logins", []))
    login_slogan = request.registry.settings.get("magiclogin.login_slogan")
    return locals()


@simple_route("/login-email", route_name="login_email", renderer='magiclogin/login_email.html', decorator=throttled_view(setting="magiclogin.login_email_throttle"))
@include_in_sitemap(False)
def login_email(request: Request):
    """Ask user email to start email sign in process."""
    schema = AskEmailSchema().bind(request=request)
    button = deform.Button(name='confirm', title="Email me a link to sign in", css_class="btn btn-primary btn-block")
    form = deform.Form(schema, buttons=[button])
    # User submitted this form
    if request.method == "POST":
        if 'confirm' in request.POST:
            try:
                appstruct = form.validate(request.POST.items())
                start_email_login(request, appstruct["email"])
                return httpexceptions.HTTPFound(request.route_url("login_email_sent"))
            except deform.ValidationFailure as e:
                # Render a form version where errors are visible next to the fields,
                # and the submitted values are posted back
                rendered_form = e.render()
        else:
            # We don't know which control caused form submission
            raise httpexceptions.HTTPInternalServerError("Unknown form button pressed")
    else:
        # Render a form with initial values
        rendered_form = form.render()

    return locals()


@simple_route("/login-email-sent", route_name="login_email_sent", renderer='magiclogin/login_email_sent.html')
@include_in_sitemap(False)
def login_email_sent(request: Request):
    email = request.session.get("email")
    return locals()


@simple_route("/verify-email-login/{token}", route_name="verify_email_login", decorator=throttled_view(setting="magiclogin.login_email_throttle"))
@include_in_sitemap(False)
def _verify_email_login(request):
    """Confirm email login token."""
    token = request.matchdict["token"]
    return verify_email_login(request, token)


@simple_route("/login-to-continue", route_name="login_to_continue", renderer='magiclogin/login_to_continue.html')
@include_in_sitemap(False)
def login_to_continue(request):
    """require_login() intersitital"""

    state = get_login_state(request)

    if state:
        msg = state.get("msg")
    else:
        # Bots banging the door
        msg = ""

    settings = request.registry.settings
    social_logins = aslist(settings.get("websauna.social_logins", []))
    login_slogan = request.registry.settings.get("magiclogin.login_slogan")
    return locals()



# Pyramid
from pyramid.httpexceptions import HTTPFound

# Websauna
from websauna.magiclogin.requirelogin import pop_login_state
from websauna.magiclogin.requirelogin import require_login
from websauna.system.core import messages
from websauna.system.core.route import simple_route


# Exposed to testing
_get_url = 0
_was_logged_in = None
_captured_post_params = None


@simple_route("/example_view_get", route_name="example_view_get", decorator=require_login())
def example_view_get(request):
    """Demostrate require_login() decorator for GET views."""
    global _get_url
    global _was_logged_in

    _get_url = request.url

    login_state = pop_login_state(request)
    params = request.params or login_state

    _was_logged_in = login_state is None

    messages.add(request, kind="info", msg="Succesfully continued GET action after login, params {}".format(params), msg_id="msg-example-get-view-success")

    return HTTPFound(request.route_url("home"))


@simple_route("/example_view_post", route_name="example_view_post", decorator=require_login())
def example_view_post(request):
    """Demostrate require_login() decorator for POST views."""
    global _was_logged_in
    global _captured_post_params

    login_state = pop_login_state(request)
    params = login_state["params"] if login_state else request.POST

    _was_logged_in = login_state is None
    _captured_post_params = params.copy()

    messages.add(request, kind="info", msg="Succesfully continued POST action after login, params {}".format(params), msg_id="msg-example-post-view-success")

    return HTTPFound(request.route_url("home"))


@simple_route("/require_login_example_page", route_name="magic_login_example_page", renderer="requirelogin-example.html")
def require_login_example_page(request):
    return locals()



"""This contains app entry point for running a demo site for this addon or running functional tests for this addon."""

# Pyramid
from pyramid.interfaces import IRequest

# Websauna
import websauna.system


class Initializer(websauna.system.DemoInitializer):
    """A demo / test app initializer for testing addon websauna.magiclogin."""

    def include_addons(self):
        """Include this addon in the configuration."""
        self.config.include("websauna.magiclogin")

    def configure_user(self):
        """Override default Websauna login system.

        We skip registration, forgot email and such functionality as it is not needed.
        """
        from websauna.magiclogin.requirelogin import DeferredActionLoginService
        from websauna.system.user import subscribers
        from websauna.system.user.credentialactivityservice import DefaultCredentialActivityService
        from websauna.system.user.interfaces import ICredentialActivityService
        from websauna.system.user.interfaces import ILoginService
        from websauna.system.user.interfaces import IRegistrationService
        from websauna.system.user.registrationservice import DefaultRegistrationService

        # Set up login service
        registry = self.config.registry
        registry.registerAdapter(factory=DeferredActionLoginService, required=(IRequest,), provided=ILoginService)
        registry.registerAdapter(factory=DefaultCredentialActivityService, required=(IRequest,), provided=ICredentialActivityService)
        registry.registerAdapter(factory=DefaultRegistrationService, required=(IRequest,), provided=IRegistrationService)

        # Set up user templates
        self.config.add_jinja2_search_path('websauna.magiclogin:templates', name='.html')
        self.config.add_jinja2_search_path('websauna.magiclogin:templates', name='.txt')
        self.config.add_jinja2_search_path('websauna.system.user:templates', name='.html')
        self.config.add_jinja2_search_path('websauna.system.user:templates', name='.txt')

        # Set up login event
        self.config.scan(subscribers)

        # Add the stock views we don't override
        self.config.add_view(view="websauna.system.user.views.login_social", route_name="login_social")
        self.config.add_route('logout', '/logout')
        self.config.add_view(view="websauna.system.user.views.logout", route_name="logout")

    def configure_static(self):
        """Configure static asset serving and cache busting."""
        super(Initializer, self).configure_static()
        self.config.registry.static_asset_policy.add_static_view('magiclogin-static', 'websauna.magiclogin:static')

    def configure_templates(self):
        super(Initializer, self).configure_templates()

        # Your app templates go here
        self.config.add_jinja2_search_path('websauna.magiclogin:demotemplates', name='.html', prepend=True)

    def configure_views(self):
        super(Initializer, self).configure_views()

        from websauna.magiclogin.tests import requireloginexamples
        self.config.scan(requireloginexamples)


def main(global_config, **settings):
    init = Initializer(global_config)
    init.run()
    return init.make_wsgi_app()



"""Passwordless login with email or SMS functionality."""

# Standard Library
import json
import logging
import random
import string
import time

# Pyramid
from pyramid.httpexceptions import HTTPFound

# Websauna
from websauna.system.core import messages
from websauna.system.core.redis import get_redis
from websauna.system.http import Request
from websauna.system.mail import send_templated_mail
from websauna.system.user.events import UserCreated
from websauna.system.user.models import User
from websauna.system.user.utils import get_login_service
from websauna.utils.time import now


#: Under which hset we store user login data
LOGIN_VERIFICATION_REDIS_HKEY = "login_verification_token"


EMAIL_TOKEN_LENGTH = 16


logger = logging.getLogger(__name__)


def rand_string(length=5):
    """http://stackoverflow.com/a/23728630/315168"""
    return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length))


def set_verification_token(request, token_type, user_id, next_url=None):
    redis = get_redis(request.registry)

    email_token_expiration_time = int(request.registry.settings.get("magiclink.email_token_expiration_seconds", 3600))

    token = rand_string(length=EMAIL_TOKEN_LENGTH)
    expires = time.time() + email_token_expiration_time

    data = {
        "token_type": token_type,
        "expires": expires,
        "token": token,
        "email": user_id,
        "next_url": next_url,
    }

    redis.hset(LOGIN_VERIFICATION_REDIS_HKEY, token, json.dumps(data))
    return token, data


def get_or_create_email_user(request: Request, email: str) -> User:
    """Fetch existing user or create new based on email."""
    dbsession = request.dbsession

    u = dbsession.query(User).filter_by(email=email).first()
    if u is not None:
        u.first_login = False
        return u

    u = User(email=email)
    u.registration_source = "email"
    u.activated_at = now()
    u.first_login = True

    dbsession.add(u)
    dbsession.flush()

    request.registry.notify(UserCreated(request, u))
    return u


def verify_email_login(request: Request, token: str):
    """Verify email login token."""

    redis = get_redis(request.registry)

    def fail(msg="Sign in link invalid. Please try again."):
        messages.add(request, kind="error", msg=msg, msg_id="msg-bad-email-token")
        return HTTPFound(request.route_url("login"))

    # Hackety hacky by our Russian friends again?
    if len(token) != EMAIL_TOKEN_LENGTH:
        logger.warn("Bad token: %s", token)
        return fail()

    token_data = redis.hget(LOGIN_VERIFICATION_REDIS_HKEY, token)
    if not token_data:
        return fail()

    # Allow use the code only once, then erase
    redis.hdel(LOGIN_VERIFICATION_REDIS_HKEY, token)

    data = json.loads(token_data.decode("utf-8"))

    # Only verify email tokens in this view
    assert data["token_type"] == "email"

    if time.time() > data["expires"]:
        return fail("Sign in link expired. Please try again.")

    email = data["email"]

    # Create new user or get existing user based on this email
    user = get_or_create_email_user(request, email)
    login_service = get_login_service(request)

    # Returns HTTPRedirect taking user to post-login page
    return login_service.authenticate_user(user, login_source="email")


def start_email_login(request: Request, email: str):
    request.session["email"] = email
    token, data = set_verification_token(request, "email", email)
    verify_link = request.route_url("verify_email_login", token=token)
    verify_minutes = int(int(request.registry.settings.get("magiclink.email_token_expiration_seconds", 300)) / 60)
    logger.info("Sending email login verification email to %s", email)
    send_templated_mail(request, [email], "magiclogin/email/verify_email", dict(verify_link=verify_link, verify_minutes=verify_minutes))



# Standard Library
import json

# Websauna
from websauna.system.core.redis import get_redis


def peek_token(redis):
    keys = redis.hkeys("login_verification_token")
    token_data = redis.hget("login_verification_token", keys[0])
    token_data = json.loads(token_data.decode("utf-8"))
    return token_data


def login(browser, test_request, web_server: str):

    # Reset login key status before proceeding
    redis = get_redis(test_request)
    redis.delete("login_verification_token")

    b = browser
    b.visit(web_server + "/login")

    assert b.is_element_present_by_css("#panel-magic-login")

    # Move to email sign in
    b.find_by_css("#nav-sign-in-by-email").click()

    b.fill("email", "foobar@example.com")
    b.find_by_css("button[name='confirm']").click()

    assert b.is_element_present_by_css("#panel-magic-login-email-sent")
    b.visit("{}/verify-email-login/{}".format(web_server, peek_token(redis)["token"]))
    assert b.is_element_present_by_css("#msg-you-are-logged-in")



# Websauna
from websauna.magiclogin.tests import requireloginexamples
from websauna.magiclogin.tests.utils import login
from websauna.magiclogin.tests.utils import peek_token
from websauna.system.core.redis import get_redis


def test_login_actions_get_email(dbsession, browser, web_server, test_request):
    """See that login actions are correctly executed when user hits example page and does GET action."""

    redis = get_redis(test_request)

    requireloginexamples._was_logged_in = None

    b = browser
    b.visit(web_server + "/require_login_example_page")

    assert b.is_element_present_by_css("#heading-require-login-example")

    b.find_by_css("#btn-get-example").click()

    assert b.is_element_present_by_css("#panel-magic-login")

    # Move to email sign in
    b.find_by_css("#nav-sign-in-by-email").click()

    b.fill("email", "foobar@example.com")
    b.find_by_css("button[name='confirm']").click()

    assert b.is_element_present_by_css("#panel-magic-login-email-sent")
    b.visit("{}/verify-email-login/{}".format(web_server, peek_token(redis)["token"]))

    # User lands after post login actions
    assert not b.is_element_present_by_css("#msg-you-are-logged-in")
    assert b.is_element_present_by_css("#msg-example-get-view-success")

    assert requireloginexamples._was_logged_in is False


def test_already_logged_actions_get(dbsession, browser, web_server, test_request):
    """Already logged in goes directly to GET page."""

    requireloginexamples._was_logged_in = None

    login(browser, test_request, web_server)

    b = browser
    b.visit(web_server + "/require_login_example_page")
    b.find_by_css("#btn-get-example").click()

    assert b.is_element_present_by_css("#msg-example-get-view-success")

    assert requireloginexamples._was_logged_in is True


def test_login_actions_post_email(dbsession, browser, web_server, test_request):
    """See that login actions are correctly executed when user hits example page and does POST action."""

    redis = get_redis(test_request)

    requireloginexamples._was_logged_in = None
    requireloginexamples._captured_post_params = None

    b = browser
    b.visit(web_server + "/require_login_example_page")

    assert b.is_element_present_by_css("#heading-require-login-example")

    b.find_by_css("#btn-post-example").click()

    assert b.is_element_present_by_css("#panel-magic-login")

    # Move to email sign in
    b.find_by_css("#nav-sign-in-by-email").click()

    b.fill("email", "foobar@example.com")
    b.find_by_css("button[name='confirm']").click()

    assert b.is_element_present_by_css("#panel-magic-login-email-sent")
    b.visit("{}/verify-email-login/{}".format(web_server, peek_token(redis)["token"]))

    # User lands after post login actions
    assert not b.is_element_present_by_css("#msg-you-are-logged-in")
    assert b.is_element_present_by_css("#msg-example-post-view-success")

    assert requireloginexamples._was_logged_in is False
    assert requireloginexamples._captured_post_params["test_value"] == "88"


def test_already_logged_actions_post(dbsession, browser, web_server, test_request):
    """Already logged in goes directly to POST page."""

    requireloginexamples._was_logged_in = None
    requireloginexamples._captured_post_params = None

    login(browser, test_request, web_server)

    b = browser
    b.visit(web_server + "/require_login_example_page")

    b.find_by_css("#btn-post-example").click()

    assert b.is_element_present_by_css("#msg-example-post-view-success")

    assert requireloginexamples._was_logged_in is True
    assert requireloginexamples._captured_post_params["test_value"] == "88"



"""An example py.test functional test case."""

# Pyramid
import transaction

# SQLAlchemy
from sqlalchemy.orm.session import Session

from splinter.driver import DriverAPI

# Websauna
from websauna.magiclogin.tests.utils import peek_token
from websauna.system.core.redis import get_redis
from websauna.system.user.models import User


def test_email_login(web_server: str, browser: DriverAPI, dbsession: Session, test_request):
    """See that we can sign up / sign in through email login."""
    # Reset login key status before proceeding
    redis = get_redis(test_request)
    redis.delete("login_verification_token")

    b = browser
    b.visit(web_server + "/login")

    assert b.is_element_present_by_css("#panel-magic-login")

    # Move to email sign in
    b.find_by_css("#nav-sign-in-by-email").click()

    b.fill("email", "foobar@example.com")
    b.find_by_css("button[name='confirm']").click()

    assert b.is_element_present_by_css("#panel-magic-login-email-sent")
    b.visit("{}/verify-email-login/{}".format(web_server, peek_token(redis)["token"]))
    assert b.is_element_present_by_css("#msg-you-are-logged-in")

    # Check we created a sane user
    with transaction.manager:
        u = dbsession.query(User).first()
        assert u.email == "foobar@example.com"
        assert u.first_login

    #
    # Do it again so we capture both new user and old user flows
    #
    b.find_by_css("#nav-logout").click()
    b.visit(web_server + "/login-email")
    b.fill("email", "foobar@example.com")
    b.find_by_css("button[name='confirm']").click()

    assert b.is_element_present_by_css("#panel-magic-login-email-sent")
    b.visit("{}/verify-email-login/{}".format(web_server, peek_token(redis)["token"]))
    assert b.is_element_present_by_css("#msg-you-are-logged-in")

    # Check the user is still sane
    with transaction.manager:
        u = dbsession.query(User).first()
        assert u.email == "foobar@example.com"
        assert not u.first_login


def test_email_login_bad_token(web_server: str, browser: DriverAPI, dbsession: Session):
    """Crap token provides meaningful user error message."""
    b = browser
    b.visit(web_server + "/verify-email-login/xxx")
    assert b.is_element_present_by_css("#msg-bad-email-token")



# Standard Library
import json
import typing as t

# Pyramid
import jinja2
from pyramid.httpexceptions import HTTPFound
from pyramid.response import Response

# Websauna
from websauna.system.core.route import get_config_route
from websauna.system.http import Request
from websauna.system.user.events import Login
from websauna.system.user.interfaces import IUser
from websauna.system.user.loginservice import DefaultLoginService


def save_login_state(request: Request, msg: t.Optional[str]=None):
    """Capture original POST/GET request before user was redirected to login page."""

    # TODO: This flattens NestedMultiDict to a normal dict
    params = dict(request.params.items())

    if not msg:
        msg = request.registry.get("websauna.magiclogin.default_proceed_login_message", "Please sign in to continue")

    # Allow raw HTML serialization
    markup = isinstance(msg, jinja2.Markup)

    saved_state = {
        "url": request.url,
        "method": request.method,
        "params": params,
        "msg": msg,
        "markup": markup,
    }
    request.session["proceed_to_login"] = json.dumps(saved_state).encode("utf-8")


def get_login_state(request: Request, remove=False):
    data = request.session.get("proceed_to_login")
    if data:

        if remove:
            del request.session["proceed_to_login"]

        data = data.decode("utf-8")

        data = json.loads(data)

        # Fix raw HTML markup
        if data.get("markup"):
            data["msg"] = jinja2.Markup(data["msg"])

        return data

    return None


def pop_login_state(request: Request):
    """Fetches the action state before login proceed interfered and marks is consumed."""
    return get_login_state(request, remove=True)


def proceed_to_login(context, request, msg):
    save_login_state(request, msg)
    return HTTPFound(request.route_url("login_to_continue"))


def require_login(msg=None):
    """Automatically prompt user to login for views that require login.

    If the view is POST endpoint, POST content is captured for the view.

    To use this decorator you need to enable custom login handler:

    :param msg: Message shown in interstitial page to the user
    """

    def outer(view: t.Callable[[object, Request], Response]):

        def inner(context, request):
            if not request.user:
                return proceed_to_login(context, request, msg=msg)
            else:
                return view(context, request)

        return inner

    return outer


class DeferredActionLoginService(DefaultLoginService):
    """Login service will perform the post-login action that was saved before login."""

    def do_post_login_actions(self, user: IUser, headers: dict, location: str=None):
        """Override what happens after login."""
        request = self.request

        self.update_login_data(user)

        e = Login(request, user)
        request.registry.notify(e)

        # Where this user was going before he or she hit login button
        login_state = get_login_state(request)
        if login_state:
            location = login_state["url"]
        else:
            # Only greet user if he/she was not continuing action
            self.greet_user(user)
            location = get_config_route(request, 'websauna.login_redirect')

        return HTTPFound(location)
